from typing import Coroutine

from cachetools import TTLCache
from limiter import Limiter


class LimitCaller:
    __slots__ = ("main_limiter", "chats", "groups")

    def __init__(self) -> None:
        """A class that controls the speed of sending requests."""

        self.main_limiter = Limiter(30)
        self.groups = TTLCache[Limiter](maxsize=9_223_372_036_854_775_807, ttl=60)
        self.chats = TTLCache[Limiter](maxsize=9_223_372_036_854_775_807, ttl=60)

    async def _call_with_limit(
        self,
        chat_id: int,
        coro: Coroutine,
        storage: TTLCache[Limiter],
        rate: int | float,
        burst: int,
    ):
        """Calls the api method

        Params:
            - chat_id: telegram chat id
            - coro: method
            - storage: chat or group storage
            - rate: call rate for limiters
            - burst: burst for limiters
        """

        async with self.main_limiter:
            limiter = storage.get(chat_id)
            if not limiter:
                limiter = Limiter(rate, burst)
                storage[chat_id] = limiter
                async with limiter:
                    return await coro
            else:
                async with limiter:
                    return await coro

    async def call(self, chat_id: int, coro: Coroutine):
        """Call the method"""

        if chat_id < 0:
            return await self._call_with_limit(chat_id, coro, self.groups, 0.32, 20)
        else:
            return await self._call_with_limit(chat_id, coro, self.chats, 0.99, 1)



from typing import Optional, TypeVar

from aiogram.methods import TelegramMethod, GetChat
from aiogram import Bot

from .limit_caller import LimitCaller

T = TypeVar("T")

class LimitedBot(Bot):

    async def _call(
        self, method: TelegramMethod[T], request_timeout: Optional[int] = None
    ):
        """
        Just to not modify __init__ method
        """
        coro = self.session(self, method, timeout=request_timeout)
        if hasattr(method, "chat_id") and not isinstance(method, GetChat):
            return await self.caller.call(method.chat_id, coro)
        else:
            return await coro

    async def __call__(
        self, method: TelegramMethod[T], request_timeout: Optional[int] = None
    ):
        caller = getattr(self, "caller", None)
        if not caller:
            caller = LimitCaller()
            self.caller = caller
            self.__call__ = LimitedBot._call
        return await LimitedBot._call(self, method, request_timeout)


def patch_bot():
    """Patches the bot, these changes are not reversible"""

    Bot.__call__ = LimitedBot.__call__



from setuptools import setup, find_packages


def readme():
    with open("README.md", "r") as f:
        return f.read()


setup(
    name="limited_aiogram",
    version="1.0.2",
    author="chazovtema",
    author_email="chazovtema@mail.ru",
    description='Limit your api calls to avoid  "Flood control exceeded"',
    long_description=readme(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=[
        "requests>=2.25.1",
        "aiogram>=3.0.0",
        "limiter>=0.3.1",
        "cachetools>=5.3.2",
    ],
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords=["aiogram", "limit"],
    project_urls={"github": "https://github.com/chazovtema/limited_aiogram"},
    python_requires=">=3.10",
)
